//
// LPTK client client.rs
// Copyright (C) 2025 Óscar García Amor <ogarcia@connectical.com>
// Distributed under terms of the GNU GPLv3 license.
//

use std::sync::Arc;

use lesspass_client::Result;
use lesspass_client::model::{NewPassword, Passwords};
use tokio::sync::RwLock;

#[derive(Debug, Clone)]
pub struct Client {
    pub client: lesspass_client::Client,
    pub username: String,
    pub password: String,
    access_token: Arc<RwLock<String>>,
    refresh_token: Arc<RwLock<String>>,
}

impl Client {
    pub fn new(url: &str, username: &str, password: &str) -> Self {
        Client {
            client: lesspass_client::Client::new(url),
            username: username.into(),
            password: password.into(),
            access_token: Arc::new(RwLock::new(String::new())),
            refresh_token: Arc::new(RwLock::new(String::new()))
        }
    }

    pub async fn read_token(&self) {
        *self.refresh_token.write().await = super::read_token().await.expect("Cannot read token from keyring");
    }

    async fn auth(&self) -> Result<()> {
        // Try use refresh token first
        let refresh_token = self.refresh_token.read().await.clone();
        let token = match self.client.refresh_token(&refresh_token).await {
            Ok(token) => token,
            // If fail perform authentication
            Err(_) => self.client.create_token(&self.username, &self.password).await?
        };
        // Save new access token in self and new refresh token in self and in keyring
        super::save_token(&token.refresh).await.expect("Cannot save token to keyring");
        *self.access_token.write().await = token.access;
        *self.refresh_token.write().await = token.refresh;
        Ok(())
    }

    pub async fn get_passwords(&self) -> Result<Passwords> {
        // Try get passwords with stored access token
        let access_token = self.access_token.read().await.clone();
        match self.client.get_passwords(&access_token).await {
            Ok(passwords) => Ok(passwords),
            Err(_) => {
                // Stored access token is not valid, refresh and retry
                self.auth().await?;
                self.client.get_passwords(&self.access_token.read().await).await
            }
        }
    }

    pub async fn save_password(&self, site: String, login: String, lowercase: bool, uppercase: bool, digits: bool, symbols: bool, length: u8, counter: u32) -> Result<()> {
        let new_password = NewPassword {
            site, login, lowercase, uppercase, symbols, digits, length, counter, version: 2
        };
        // Try save password with stored access token
        let access_token = self.access_token.read().await.clone();
        match self.client.post_password(&access_token, &new_password).await {
            Ok(()) => Ok(()),
            Err(_) => {
                // Stored access token is not valid, refresh and retry
                self.auth().await?;
                self.client.post_password(&self.access_token.read().await, &new_password).await
            }
        }
    }

    pub async fn update_password(&self, id: String, site: String, login: String, lowercase: bool, uppercase: bool, digits: bool, symbols: bool, length: u8, counter: u32) -> Result<()> {
        let new_password = NewPassword {
            site, login, lowercase, uppercase, symbols, digits, length, counter, version: 2
        };
        // Try update password with stored access token
        let access_token = self.access_token.read().await.clone();
        match self.client.put_password(&access_token, &id, &new_password).await {
            Ok(()) => Ok(()),
            Err(_) => {
                // Stored access token is not valid, refresh and retry
                self.auth().await?;
                self.client.put_password(&self.access_token.read().await, &id, &new_password).await
            }
        }
    }

    pub async fn delete_password(&self, id: String) -> Result<()> {
        // Try delete password with stored access token
        let access_token = self.access_token.read().await.clone();
        match self.client.delete_password(&access_token, &id).await {
            Ok(()) => Ok(()),
            Err(_) => {
                // Stored access token is not valid, refresh and retry
                self.auth().await?;
                self.client.delete_password(&self.access_token.read().await, &id).await
            }
        }
    }
}

impl Default for Client {
    fn default() -> Self {
        Self::new("", "", "")
    }
}
