//
// LPTK client secret.rs
// Copyright (C) 2025 Óscar García Amor <ogarcia@connectical.com>
// Distributed under terms of the GNU GPLv3 license.
//


use std::cell::RefCell;
use std::collections::HashMap;

#[derive(Debug, Default)]
pub struct Secrets {
    pub server: RefCell<String>,
    pub username: RefCell<String>,
    pub password: RefCell<String>
}

impl Secrets {
    pub fn new() -> Secrets {
        Secrets::default()
    }

    async fn read() -> oo7::Result<(String, String, String)> {
        // Read credentials from Keyring
        let keyring = oo7::Keyring::new().await?;
        // Search by LPTK secret
        let attributes = HashMap::from([("LPTK", "secret")]);
        let keys = keyring.search_items(&attributes).await?;
        if keys.is_empty() {
            Ok((String::new(), String::new(), String::new()))
        } else {
            // Get password from secret and atributes
            let password = keys[0].secret().await?;
            let password = String::from_utf8(password.to_vec()).expect("Could not decode password");
            let attributes = keys[0].attributes().await?;
            let server = attributes.get("server").unwrap_or(&String::new()).clone();
            let username = attributes.get("username").unwrap_or(&String::new()).clone();
            Ok((server, username, password))
        }
    }

    pub async fn update(&self) {
        let secrets = Self::read().await.expect("Could not obtain secrets");
        *self.server.borrow_mut() = secrets.0;
        *self.username.borrow_mut() = secrets.1;
        *self.password.borrow_mut() = secrets.2;
    }

    pub async fn save(&self) -> oo7::Result<()> {
        let keyring = oo7::Keyring::new().await?;
        let server = self.server.borrow();
        let username = self.username.borrow();
        let password = self.password.borrow();
        // Search by LPTK secret to known if has previous credentials stored
        let search_attributes = HashMap::from([("LPTK", "secret")]);
        let keys = keyring.search_items(&search_attributes).await?;
        // Set the new attributes
        let attributes: HashMap<&str, &str> = HashMap::from([("LPTK", "secret"), ("server", &server), ("username", &username)]);
        if ! keys.is_empty() {
            // Delete previous credentials before store new ones
            keyring.delete(&search_attributes).await?;
        }
        // Store the new credentials
        keyring.create_item("LPTK: server credentials", &attributes, password.as_str(), true).await?;
        Ok(())
    }
}

pub async fn read_token() -> oo7::Result<String> {
    // Read token from Keyring
    let keyring = oo7::Keyring::new().await?;
    // Search by LPTK token
    let attributes = HashMap::from([("LPTK", "token")]);
    let keys = keyring.search_items(&attributes).await?;
    if keys.is_empty() {
        Ok(String::new())
    } else {
        // Get token from secret
        let token = keys[0].secret().await?;
        let token = String::from_utf8(token.to_vec()).expect("Could not decode password");
        Ok(token)
    }
}

pub async fn save_token(token: &str) -> oo7::Result<()> {
    let keyring = oo7::Keyring::new().await?;
    // Search by LPTK token to known if has previous credentials stored
    let attributes = HashMap::from([("LPTK", "token")]);
    let keys = keyring.search_items(&attributes).await?;
    if ! keys.is_empty() {
        // Delete previous credentials before store new ones
        keyring.delete(&attributes).await?;
    }
    // Store the new token
    keyring.create_item("LPTK: server token", &attributes, token, true).await?;
    Ok(())
}
