//
// LPTK ui generator.rs
// Copyright (C) 2025 Óscar García Amor <ogarcia@connectical.com>
// Distributed under terms of the GNU GPLv3 license.
//
use std::cell::{Cell, OnceCell, RefCell};
use std::sync::OnceLock;

use gettextrs::gettext;
use glib::clone;
use glib::subclass::Signal;
use adw::prelude::*;
use adw::subclass::prelude::*;
use gtk::{gio, glib};

use crate::config::APP_ID;
use crate::model::Password;

mod imp {
    use super::*;

    #[derive(Debug, Default, gtk::CompositeTemplate)]
    #[template(resource = "/me/ogarcia/lptk/gtk/generator.ui")]
    pub struct FragmentGenerator {
        // Template widgets
        #[template_child]
        pub entry_site: TemplateChild<gtk::Entry>,
        #[template_child]
        pub entry_user: TemplateChild<gtk::Entry>,
        #[template_child]
        pub entry_master_password: TemplateChild<gtk::PasswordEntry>,
        #[template_child]
        pub checkbox_lower: TemplateChild<gtk::CheckButton>,
        #[template_child]
        pub checkbox_upper: TemplateChild<gtk::CheckButton>,
        #[template_child]
        pub checkbox_digits: TemplateChild<gtk::CheckButton>,
        #[template_child]
        pub checkbox_symbols: TemplateChild<gtk::CheckButton>,
        #[template_child]
        pub spin_length: TemplateChild<gtk::SpinButton>,
        #[template_child]
        pub spin_counter: TemplateChild<gtk::SpinButton>,
        #[template_child]
        pub button_grid: TemplateChild<gtk::FlowBox>,
        #[template_child]
        pub button_generate: TemplateChild<gtk::Button>,
        #[template_child]
        pub button_save: TemplateChild<gtk::Button>,
        #[template_child]
        pub button_delete: TemplateChild<gtk::Button>,
        #[template_child]
        pub button_clear: TemplateChild<gtk::Button>,
        #[template_child]
        pub entry_generated_password: TemplateChild<gtk::Entry>,
        pub generated_password: RefCell<String>,
        pub generated_password_revealed: Cell<bool>,
        pub password: RefCell<Password>,
        pub settings: OnceCell<gio::Settings>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for FragmentGenerator {
        const NAME: &'static str = "FragmentGenerator";
        type Type = super::FragmentGenerator;
        type ParentType = gtk::Box;

        fn class_init(klass: &mut Self::Class) {
            klass.bind_template();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    impl ObjectImpl for FragmentGenerator {
        fn signals() -> &'static [Signal] {
            // Setup a signal to call clear action
            static SIGNALS: OnceLock<Vec<Signal>> = OnceLock::new();
            SIGNALS.get_or_init(|| {
                vec![Signal::builder("clear").param_types([bool::static_type()]).build(),
                    Signal::builder("toast").param_types([String::static_type()]).build(),
                    Signal::builder("visible-save-button").param_types([bool::static_type()]).build(),
                    Signal::builder("enable-save-button").param_types([bool::static_type()]).build(),
                    Signal::builder("save-action").param_types([bool::static_type()]).build(),
                    Signal::builder("delete-action").param_types([bool::static_type()]).build(),
                    Signal::builder("password-save").param_types([Password::static_type()]).build(),
                    Signal::builder("password-delete").param_types([String::static_type()]).build()]
            })
        }

        fn constructed(&self) {
            // Call "constructed" on parent
            self.parent_constructed();

            // Hide generated password
            self.entry_generated_password.set_visible(false);

            // Setup
            let obj = self.obj();
            obj.setup_settings();
            obj.setup_callbacks();
            obj.setup_signals();
        }
    }

    impl WidgetImpl for FragmentGenerator {}
    impl BoxImpl for FragmentGenerator {}
}

glib::wrapper! {
    pub struct FragmentGenerator(ObjectSubclass<imp::FragmentGenerator>)
        @extends gtk::Box, gtk::Widget,
        @implements gtk::Accessible, gtk::Buildable, gtk::ConstraintTarget, gtk::Orientable;
}

impl FragmentGenerator {
    pub fn new() -> Self {
        glib::Object::builder().build()
    }

    pub fn load_password(&self, password: Password) {
        self.imp().entry_site.set_text(&password.property::<String>("site"));
        self.imp().entry_user.set_text(&password.property::<String>("login"));
        self.imp().entry_master_password.set_text("");
        self.imp().checkbox_lower.set_active(password.property::<bool>("lower"));
        self.imp().checkbox_upper.set_active(password.property::<bool>("upper"));
        self.imp().checkbox_digits.set_active(password.property::<bool>("digits"));
        self.imp().checkbox_symbols.set_active(password.property::<bool>("symbols"));
        self.imp().spin_length.set_value(password.property::<u8>("length") as f64);
        self.imp().spin_counter.set_value(password.property::<u32>("counter") as f64);
        *self.imp().password.borrow_mut() = password;
        self.imp().entry_generated_password.set_visible(false);
        self.imp().button_save.set_label(&gettext("Update"));
        self.imp().button_grid.action_set_enabled("generator.save", false);
        self.imp().button_grid.action_set_enabled("generator.delete", true);
        self.imp().button_delete.set_visible(true);
        self.imp().entry_master_password.grab_focus();
    }

    fn setup_settings(&self) {
        // Read settings from application schema
        let settings = gio::Settings::new(APP_ID);
        self.imp().settings.set(settings).expect("Could not read settings");
    }

    #[inline(always)]
    fn settings(&self) -> &gio::Settings {
        // Return settings from self
        self.imp().settings.get().expect("Could not obtain settings")
    }

    fn can_generate_can_save(&self) -> (bool, bool) {
        // Check if user can generate / can save/update a password
        let site = self.imp().entry_site.text();
        let user = self.imp().entry_user.text();
        let master_password = self.imp().entry_master_password.text();
        let lower = self.imp().checkbox_lower.is_active();
        let upper = self.imp().checkbox_upper.is_active();
        let digits = self.imp().checkbox_digits.is_active();
        let symbols = self.imp().checkbox_symbols.is_active();
        let length = self.imp().spin_length.value() as u8;
        let counter = self.imp().spin_counter.value() as u32;
        let can_generate = if site == "" || user == "" || master_password == "" || (!lower && !upper && !digits && !symbols) {
            false
        } else {
            true
        };
        let can_save = if site == "" || user == "" || (!lower && !upper && !digits && !symbols) {
            false
        } else {
            let password = self.imp().password.borrow().clone();
            let stored_site = password.property::<String>("site");
            let stored_user = password.property::<String>("login");
            let stored_lower = password.property::<bool>("lower");
            let stored_upper = password.property::<bool>("upper");
            let stored_digits = password.property::<bool>("digits");
            let stored_symbols = password.property::<bool>("symbols");
            let stored_length = password.property::<u8>("length");
            let stored_counter = password.property::<u32>("counter");
            if site != stored_site || user != stored_user || stored_lower != lower ||
                stored_upper != upper || stored_digits != digits || stored_symbols != symbols ||
                stored_length != length || stored_counter != counter {
                    true
            } else {
                false
            }
        };
        (can_generate, can_save)
    }

    fn setup_callbacks(&self) {
        // Create an action group and register into button grid
        let button_grid_action_group = gio::SimpleActionGroup::new();
        self.imp().button_grid.insert_action_group("generator", Some(&button_grid_action_group));

        // Create button actions
        let action_generate = gio::SimpleAction::new("generate", None);
        let action_clear = gio::SimpleAction::new("clear", None);
        let action_save = gio::SimpleAction::new("save", None);
        let action_delete = gio::SimpleAction::new("delete", None);

        // Define generate action
        action_generate.connect_activate(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, _| {
                let site = window.imp().entry_site.text();
                let user = window.imp().entry_user.text();
                let master_password = window.imp().entry_master_password.text();
                let lower = window.imp().checkbox_lower.is_active();
                let upper = window.imp().checkbox_upper.is_active();
                let digits = window.imp().checkbox_digits.is_active();
                let symbols = window.imp().checkbox_symbols.is_active();
                let length = window.imp().spin_length.value() as u8;
                let counter = window.imp().spin_counter.value() as u16;
                // Generate password with user entry
                let mut charset = lesspass::CharacterSet::All;
                if ! lower { charset.remove(lesspass::CharacterSet::Lowercase); }
                if ! upper { charset.remove(lesspass::CharacterSet::Uppercase); }
                if ! digits { charset.remove(lesspass::CharacterSet::Numbers); }
                if ! symbols { charset.remove(lesspass::CharacterSet::Symbols); }
                let salt = lesspass::generate_salt(&site, &user, counter.into());
                let entropy = lesspass::generate_entropy(&master_password, &salt, lesspass::Algorithm::SHA256, 100000);
                let generated_password = lesspass::render_password(&entropy, charset, length.into());
                // Copy to clipboard and show box to user
                window.clipboard().set_text(&generated_password);
                window.imp().entry_generated_password.set_text(&"●".repeat(generated_password.len()));
                window.imp().entry_generated_password.set_visible(true);
                *window.imp().generated_password.borrow_mut() = generated_password;
                window.imp().generated_password_revealed.set(false);
            }
        ));

        // Define save action
        action_save.connect_activate(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, _| {
                let password = window.imp().password.borrow().clone();
                let id = password.property::<String>("id");
                let site = window.imp().entry_site.text();
                let user = window.imp().entry_user.text();
                let lower = window.imp().checkbox_lower.is_active();
                let upper = window.imp().checkbox_upper.is_active();
                let digits = window.imp().checkbox_digits.is_active();
                let symbols = window.imp().checkbox_symbols.is_active();
                let length = window.imp().spin_length.value() as u8;
                let counter = window.imp().spin_counter.value() as u32;
                let new_password = Password::new(
                    id, site.to_string(), user.to_string(),
                    lower, upper, digits, symbols, length, counter
                );
                window.emit_by_name::<()>("password-save", &[&new_password]);
                *window.imp().password.borrow_mut() = new_password;
                window.imp().button_grid.action_set_enabled("generator.save", false);
            }
        ));

        // Define delete action
        action_delete.connect_activate(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, _| {
                let password = window.imp().password.borrow().clone();
                let id = password.property::<String>("id");
                let site = password.property::<String>("site");
                let login = password.property::<String>("login");
                let dialog = adw::AlertDialog::builder()
                    .heading(&gettext("Delete password"))
                    .body(&gettext(format!("Are you sure you want to delete the profile `{site} / {login}`?")))
                    .default_response("no")
                    .build();
                dialog.add_response("no", &gettext("Nope"));
                dialog.add_response("yes", &gettext("Yes"));
                dialog.set_response_appearance("yes", adw::ResponseAppearance::Destructive);
                dialog.present(Some(&window));

                dialog.connect_response(None, move |_, response| {
                    if response == "yes" {
                        window.emit_by_name::<()>("password-delete", &[&id]);
                    }
                });
            }
        ));

        // Define clear action
        action_clear.connect_activate(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, _| {
                *window.imp().password.borrow_mut() = Password::default();
                window.imp().entry_site.set_text("");
                window.imp().entry_user.set_text(&window.settings().string("user"));
                window.imp().entry_master_password.set_text("");
                window.imp().checkbox_lower.set_active(window.settings().boolean("lower"));
                window.imp().checkbox_upper.set_active(window.settings().boolean("upper"));
                window.imp().checkbox_digits.set_active(window.settings().boolean("digits"));
                window.imp().checkbox_symbols.set_active(window.settings().boolean("symbols"));
                window.imp().spin_length.set_value(window.settings().double("length"));
                window.imp().spin_counter.set_value(window.settings().double("counter"));
                window.imp().button_save.set_label(&gettext("Save"));
                window.imp().button_grid.action_set_enabled("generator.delete", false);
                window.imp().button_delete.set_visible(false);
                window.imp().entry_generated_password.set_visible(false)
            }
        ));

        // Add actions to group
        button_grid_action_group.add_action(&action_clear);
        button_grid_action_group.add_action(&action_generate);
        button_grid_action_group.add_action(&action_save);
        button_grid_action_group.add_action(&action_delete);

        // Call to clear action to load default password settings
        button_grid_action_group.activate_action("clear", None);

        // Disable generate and save buttons until the user fills in all entry boxes
        action_generate.set_enabled(false);
        action_save.set_enabled(false);

        // Checks entry boxes to enable or disable buttons
        let enable_buttons = clone!(
            #[weak(rename_to = window)]
            self,
            move || {
                let (can_generate, can_save) = window.can_generate_can_save();
                action_generate.set_enabled(can_generate);
                window.imp().button_grid.action_set_enabled("generator.save", can_save);
                action_save.set_enabled(can_save);
            }
        );

        // Connect entry boxes with enable checker
        self.imp().entry_site.connect_changed(clone!(#[strong] enable_buttons, move |_| enable_buttons()));
        self.imp().entry_user.connect_changed(clone!(#[strong] enable_buttons, move |_| enable_buttons()));
        self.imp().entry_master_password.connect_changed(clone!(#[strong] enable_buttons, move |_| enable_buttons()));

        // Connect check buttons with enable checker
        self.imp().checkbox_lower.connect_toggled(clone!(#[strong] enable_buttons, move |_| enable_buttons()));
        self.imp().checkbox_upper.connect_toggled(clone!(#[strong] enable_buttons, move |_| enable_buttons()));
        self.imp().checkbox_digits.connect_toggled(clone!(#[strong] enable_buttons, move |_| enable_buttons()));
        self.imp().checkbox_symbols.connect_toggled(clone!(#[strong] enable_buttons, move |_| enable_buttons()));

        // Connect spin buttons with with enable checker
        self.imp().spin_length.connect_value_changed(clone!(#[strong] enable_buttons, move |_| enable_buttons()));
        self.imp().spin_counter.connect_value_changed(clone!(#[strong] enable_buttons, move |_| enable_buttons()));

        // Checks if all entry boxes are filled to perform generate action when user press intro in any entry box
        let perform_generate = clone!(
            #[weak(rename_to = window)]
            self,
            move || {
                let entry_site_text = window.imp().entry_site.text();
                let entry_user_text = window.imp().entry_user.text();
                let entry_master_password_text = window.imp().entry_master_password.text();
                if entry_site_text != "" && entry_user_text != "" && entry_master_password_text != "" {
                    button_grid_action_group.activate_action("generate", None)
                }
            }
        );

        // Connect entry boxes with perform checker
        self.imp().entry_site.connect_activate(clone!(#[strong] perform_generate, move |_| perform_generate()));
        self.imp().entry_user.connect_activate(clone!(#[strong] perform_generate, move |_| perform_generate()));
        self.imp().entry_master_password.connect_activate(clone!(#[strong] perform_generate, move |_| perform_generate()));

        // Connect spin boxes with perform checker
        self.imp().spin_length.connect_activate(clone!(#[strong] perform_generate, move |_| perform_generate()));
        self.imp().spin_counter.connect_activate(clone!(#[strong] perform_generate, move |_| perform_generate()));

        // Generated password buttons actions
        self.imp().entry_generated_password.connect_icon_press(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, icon| {
                let generated_password = window.imp().generated_password.borrow();
                match icon {
                    gtk::EntryIconPosition::Primary => {
                        // Copy button, copy generated password to clipboard
                        window.clipboard().set_text(&generated_password);
                        // Emit a toast
                        window.emit_by_name::<()>("toast", &[&gettext("Password copied to clipboard")]);
                    },
                    _ => {
                        // Reveal button
                        if window.imp().generated_password_revealed.get() {
                            // The password is being displayed, hidden it
                            window.imp().entry_generated_password.set_text(&"●".repeat(generated_password.len()));
                            window.imp().generated_password_revealed.set(false);
                        } else {
                            // The password is not being displayed, show it
                            window.imp().entry_generated_password.set_text(&generated_password);
                            window.imp().generated_password_revealed.set(true);
                        }
                    }
                }
            }
        ));
    }

    fn setup_signals(&self) {
        // Listen for clear signal
        self.connect_closure("clear", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: super::FragmentGenerator, _: bool| {
                window.imp().button_grid.activate_action("generator.clear", None).expect("Could not clear generator");
            }
        ));

        // Listen for save button visibility signal
        self.connect_closure("visible-save-button", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: super::FragmentGenerator, visible: bool| {
                window.imp().button_save.set_visible(visible);
            }
        ));

        // Listen for enable save button signal
        self.connect_closure("enable-save-button", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: super::FragmentGenerator, enable: bool| {
                window.imp().button_grid.action_set_enabled("generator.save", enable);
            }
        ));

        // Listen for save action signal
        self.connect_closure("save-action", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: super::FragmentGenerator, _: bool| {
                let (_, can_save) = window.can_generate_can_save();
                if can_save {
                    window.imp().button_grid.activate_action("generator.save", None).expect("Could not save password");
                }
            }
        ));

        // Listen for delete action signal
        self.connect_closure("delete-action", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: super::FragmentGenerator, _: bool| {
                let password = window.imp().password.borrow().clone();
                let id = password.property::<String>("id");
                if ! id.is_empty() {
                    // If has stored id then can be deleted
                    window.imp().button_grid.activate_action("generator.delete", None).expect("Could not delete password");
                }
            }
        ));
    }
}
