//
// LPTK ui password_list.rs
// Copyright (C) 2025 Óscar García Amor <ogarcia@connectical.com>
// Distributed under terms of the GNU GPLv3 license.
//

use std::cell::RefCell;
use std::sync::OnceLock;

use gettextrs::gettext;
use glib::clone;
use gtk::prelude::*;
use adw::subclass::prelude::*;
use glib::subclass::Signal;
use gtk::{gio, glib};

use crate::model::Password;
use super::FragmentPassword;

mod imp {
    use super::*;

    #[derive(Debug, Default, gtk::CompositeTemplate)]
    #[template(resource = "/me/ogarcia/lptk/gtk/password-list.ui")]
    pub struct FragmentPasswordList {
        // Template widgets
        #[template_child]
        pub configure_first: TemplateChild<adw::StatusPage>,
        #[template_child]
        pub content: TemplateChild<adw::Clamp>,
        #[template_child]
        pub search_entry: TemplateChild<gtk::SearchEntry>,
        #[template_child]
        pub password_list: TemplateChild<gtk::ListView>,
        pub passwords: RefCell<Option<gio::ListStore>>,
        pub filter_model: RefCell<Option<gtk::FilterListModel>>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for FragmentPasswordList {
        const NAME: &'static str = "FragmentPasswordList";
        type Type = super::FragmentPasswordList;
        type ParentType = gtk::Box;

        fn class_init(klass: &mut Self::Class) {
            klass.bind_template();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    impl ObjectImpl for FragmentPasswordList {
        fn signals() -> &'static [Signal] {
            // Setup a signal that is emitted when the user selects a password from list
            static SIGNALS: OnceLock<Vec<Signal>> = OnceLock::new();
            SIGNALS.get_or_init(|| {
                vec![Signal::builder("password-selected").param_types([Password::static_type()]).build()]
            })
        }

        fn constructed(&self) {
            // Call "constructed" on parent
            self.parent_constructed();

            // Load settings and setup callbacks
            let obj = self.obj();
            obj.setup_factory();
            obj.setup_callbacks();
        }
    }

    impl WidgetImpl for FragmentPasswordList {}
    impl BoxImpl for FragmentPasswordList {}
}

glib::wrapper! {
    pub struct FragmentPasswordList(ObjectSubclass<imp::FragmentPasswordList>)
        @extends gtk::Box, gtk::Widget,
        @implements gtk::Accessible, gtk::Buildable, gtk::ConstraintTarget, gtk::Orientable;
}

impl FragmentPasswordList {
    pub fn new() -> Self {
        glib::Object::builder().build()
    }

    fn passwords(&self) -> gio::ListStore {
        self.imp().passwords.borrow().clone().expect("Could not get password list store")
    }

    pub fn clear(&self) {
        self.imp().passwords.replace(None);
        self.imp().content.set_visible(false);
        self.imp().configure_first.set_visible(true);
    }

    pub fn populate(&self, password_list: lesspass_client::model::Passwords) {
        // Internal function to compare passwords and sort it
        let cmp_func = |a: &glib::Object, b: &glib::Object| {
            let order_a: String = a.property("site");
            let order_b: String = b.property("site");
            match order_a.cmp(&order_b) {
                std::cmp::Ordering::Equal => {
                    let login_a: String = a.property("login");
                    let login_b: String = b.property("login");
                    login_a.cmp(&login_b)
                },
                other => other
            }
        };
        if password_list.results.is_empty() {
            self.imp().configure_first.set_icon_name(Some("view-list-symbolic"));
            self.imp().configure_first.set_title(&gettext("No data"));
            self.imp().configure_first.set_description(Some(&gettext("The password list is empty, add some first and you can see them here")));
        } else {
            self.imp().configure_first.set_visible(false);
            self.imp().content.set_visible(true);
            let model = gio::ListStore::new::<Password>();
            self.imp().passwords.replace(Some(model));
            for password in password_list.results {
                let p = Password::new(password.id,
                    password.site,
                    password.login,
                    password.lowercase,
                    password.uppercase,
                    password.digits,
                    password.symbols,
                    password.length,
                    password.counter);
                self.passwords().insert_sorted(&p, &cmp_func);
            }
            let filter_model = gtk::FilterListModel::new(Some(self.passwords()), None::<gtk::CustomFilter>);
            self.imp().filter_model.replace(Some(filter_model.clone()));
            let selection_model = gtk::SingleSelection::builder()
                .autoselect(false)
                .model(&filter_model)
                .build();
            self.imp().password_list.set_model(Some(&selection_model));
        }
    }

    fn setup_factory(&self) {
        let factory = gtk::SignalListItemFactory::new();

        factory.connect_setup(move |_, list_item| {
            let password_row = FragmentPassword::new();
            list_item
                .downcast_ref::<gtk::ListItem>()
                .expect("Needs to be ListItem")
                .set_child(Some(&password_row));
        });

        factory.connect_bind(move |_, list_item| {
            // Get Password from ListItem
            let password = list_item
                .downcast_ref::<gtk::ListItem>()
                .expect("Needs to be ListItem")
                .item()
                .and_downcast::<Password>()
                .expect("The item has to be an Password");
            // Get FragmentPassword (password row) from ListItem
            let password_row = list_item
                .downcast_ref::<gtk::ListItem>()
                .expect("Needs to be ListItem")
                .child()
                .and_downcast::<FragmentPassword>()
                .expect("The child has to be a FragmentPassword");
            password_row.bind(&password);
        });

        // Tell factory how to unbind FragmentPassword (password row) from Password
        factory.connect_unbind(move |_, list_item| {
            // Get FragmentPassword (password row) from ListItem
            let password_row = list_item
                .downcast_ref::<gtk::ListItem>()
                .expect("Needs to be ListItem")
                .child()
                .and_downcast::<FragmentPassword>()
                .expect("The child has to be a `FragmentPassword");
            password_row.unbind();
        });

        // Set the factory of the list view
        self.imp().password_list.set_factory(Some(&factory));
    }

    fn setup_callbacks(&self) {
        let gesture_click = gtk::GestureClick::new();
        gesture_click.set_button(gtk::gdk::ffi::GDK_BUTTON_PRIMARY as u32);
        gesture_click.connect_pressed(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, _, _, _| {
                if let Some(selection_model) = window.imp().password_list.model().and_then(|m| m.downcast::<gtk::SingleSelection>().ok()) {
                    let selected_index = selection_model.selected();
                    if let Some(item) = selection_model.model().and_then(|m| m.item(selected_index)).and_then(|o| o.downcast::<Password>().ok()) {
                        // Send a signal with selected password to main window
                        window.emit_by_name::<()>("password-selected", &[&item]);
                    }
                }
            }
        ));
        self.imp().password_list.add_controller(gesture_click);

        self.imp().search_entry.connect_changed(clone!(
            #[weak(rename_to = window)]
            self,
            move |entry| {
                let text = entry.text().to_lowercase();
                if let Some(filter_model) = window.imp().filter_model.borrow().as_ref() {
                    filter_model.set_filter(Some(&gtk::CustomFilter::new(move |item| {
                        let password = item.downcast_ref::<Password>().expect("The item should be a Password");
                        let site = password.site().to_lowercase();
                        let login = password.login().to_lowercase();
                        site.contains(&text) || login.contains(&text)
                    })));
                };
            }
        ));
    }
}
