//
// LPTK ui preferences.rs
// Copyright (C) 2025 Óscar García Amor <ogarcia@connectical.com>
// Distributed under terms of the GNU GPLv3 license.
//

use std::cell::{Cell, OnceCell};
use std::sync::OnceLock;

use gettextrs::gettext;
use adw::prelude::*;
use adw::subclass::prelude::*;
use glib::clone;
use glib::subclass::Signal;
use gtk::{gio, glib};

use crate::client::Secrets;
use crate::config::APP_ID;

mod imp {
    use super::*;

    #[derive(Debug, Default, gtk::CompositeTemplate)]
    #[template(resource = "/me/ogarcia/lptk/gtk/preferences.ui")]
    pub struct Preferences {
        #[template_child]
        pub user: TemplateChild<adw::EntryRow>,
        #[template_child]
        pub lower: TemplateChild<adw::SwitchRow>,
        #[template_child]
        pub upper: TemplateChild<adw::SwitchRow>,
        #[template_child]
        pub digits: TemplateChild<adw::SwitchRow>,
        #[template_child]
        pub symbols: TemplateChild<adw::SwitchRow>,
        #[template_child]
        pub length: TemplateChild<adw::SpinRow>,
        #[template_child]
        pub counter: TemplateChild<adw::SpinRow>,
        #[template_child]
        pub server: TemplateChild<adw::EntryRow>,
        #[template_child]
        pub username: TemplateChild<adw::EntryRow>,
        #[template_child]
        pub password: TemplateChild<adw::PasswordEntryRow>,
        #[template_child]
        pub save: TemplateChild<adw::ButtonRow>,
        pub settings: OnceCell<gio::Settings>,
        pub secrets: OnceCell<Secrets>,
        pub secrets_changed: Cell<bool>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for Preferences {
        const NAME: &'static str = "Preferences";
        type Type = super::Preferences;
        type ParentType = adw::PreferencesDialog;

        fn class_init(klass: &mut Self::Class) {
            klass.bind_template();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    impl ObjectImpl for Preferences {
        fn signals() -> &'static [Signal] {
            // Setup a signal that is emitted when the window closes
            static SIGNALS: OnceLock<Vec<Signal>> = OnceLock::new();
            SIGNALS.get_or_init(|| {
                vec![Signal::builder("close").param_types([bool::static_type()]).build()]
            })
        }

        fn constructed(&self) {
            // Call "constructed" on parent
            self.parent_constructed();

            // Load settings and setup callbacks
            let obj = self.obj();
            obj.setup_settings();
            obj.load_preferences();
            obj.setup_callbacks();

            // Emit close signal
            obj.connect_destroy(|obj| {
                obj.emit_by_name::<()>("close", &[&obj.imp().secrets_changed.get()]);
            });
        }
    }

    impl WidgetImpl for Preferences {}
    impl AdwDialogImpl for Preferences {}
    impl PreferencesDialogImpl for Preferences {}
}

glib::wrapper! {
    pub struct Preferences(ObjectSubclass<imp::Preferences>)
        @extends adw::Dialog, adw::PreferencesDialog, gtk::Widget;
}

impl Preferences {
    pub fn new() -> Self {
        glib::Object::builder().build()
    }

    fn setup_settings(&self) {
        // Read settings from application schema
        let settings = gio::Settings::new(APP_ID);
        self.imp().settings.set(settings).expect("Could not read settings");
    }

    #[inline(always)]
    fn settings(&self) -> &gio::Settings {
        // Return settings from self
        self.imp().settings.get().expect("Could not obtain settings")
    }

    fn load_preferences(&self) {
        // Get preferences from settings
        self.imp().user.set_text(&self.settings().string("user"));
        self.imp().lower.set_active(self.settings().boolean("lower"));
        self.imp().upper.set_active(self.settings().boolean("upper"));
        self.imp().digits.set_active(self.settings().boolean("digits"));
        self.imp().symbols.set_active(self.settings().boolean("symbols"));
        self.imp().length.set_value(self.settings().double("length"));
        self.imp().counter.set_value(self.settings().double("counter"));
        glib::spawn_future_local(clone!(
            #[weak(rename_to = window)]
            self,
            async move {
                let secrets = Secrets::new();
                secrets.update().await;
                window.imp().server.set_text(&secrets.server.borrow());
                window.imp().username.set_text(&secrets.username.borrow());
                window.imp().secrets.get_or_init(|| secrets);
            }
        ));
    }

    fn setup_callbacks(&self) {
        self.imp().user.connect_text_notify(clone!(
            #[weak(rename_to = window)]
            self,
            move |entry| {
                window.settings().set_string("user", entry.text().as_str())
                    .expect("Could not set user setting");
            }
        ));

        self.imp().lower.connect_active_notify(clone!(
            #[weak(rename_to = window)]
            self,
            move |switch| {
                // Cannot deactivate all switches
                if ! window.imp().lower.is_active() && ! window.imp().upper.is_active() && ! window.imp().digits.is_active() && ! window.imp().symbols.is_active() {
                    window.imp().lower.set_active(true);
                }
                window.settings().set_boolean("lower", switch.is_active())
                    .expect("Could not set lower setting");
            }
        ));

        self.imp().upper.connect_active_notify(clone!(
            #[weak(rename_to = window)]
            self,
            move |switch| {
                // Cannot deactivate all switches
                if ! window.imp().lower.is_active() && ! window.imp().upper.is_active() && ! window.imp().digits.is_active() && ! window.imp().symbols.is_active() {
                    window.imp().upper.set_active(true);
                }
                window.settings().set_boolean("upper", switch.is_active())
                    .expect("Could not set upper setting");
            }
        ));

        self.imp().digits.connect_active_notify(clone!(
            #[weak(rename_to = window)]
            self,
            move |switch| {
                // Cannot deactivate all switches
                if ! window.imp().lower.is_active() && ! window.imp().upper.is_active() && ! window.imp().digits.is_active() && ! window.imp().symbols.is_active() {
                    window.imp().digits.set_active(true);
                }
                window.settings().set_boolean("digits", switch.is_active())
                    .expect("Could not set digits setting");
            }
        ));

        self.imp().symbols.connect_active_notify(clone!(
            #[weak(rename_to = window)]
            self,
            move |switch| {
                // Cannot deactivate all switches
                if ! window.imp().lower.is_active() && ! window.imp().upper.is_active() && ! window.imp().digits.is_active() && ! window.imp().symbols.is_active() {
                    window.imp().symbols.set_active(true);
                }
                window.settings().set_boolean("symbols", switch.is_active())
                    .expect("Could not set symbols setting");
            }
        ));

        self.imp().length.connect_value_notify(clone!(
            #[weak(rename_to = window)]
            self,
            move |spin| {
                window.settings().set_double("length", spin.value())
                    .expect("Could not set length setting");
            }
        ));

        self.imp().counter.connect_value_notify(clone!(
            #[weak(rename_to = window)]
            self,
            move |spin| {
                window.settings().set_double("counter", spin.value())
                    .expect("Could not set counter setting");
            }
        ));

        self.imp().save.connect_activated(clone!(
            #[weak(rename_to = window)]
            self,
            move |_| {
                glib::spawn_future_local(clone!(
                    #[strong]
                    window,
                    async move {
                        let secrets = window.imp().secrets.get().expect("Cannot obtain secret store");
                        let server = window.imp().server.text();
                        let username = window.imp().username.text();
                        let password = window.imp().password.text();
                        let mut secrets_changed = false;
                        if server != *secrets.server.borrow() {
                            *secrets.server.borrow_mut() = server.to_string();
                            secrets_changed = true;
                        }
                        if username != *secrets.username.borrow() {
                            if password.is_empty() {
                                let toast = adw::Toast::new(&gettext("You must retype your password to change your user name"));
                                window.add_toast(toast);
                            } else {
                                *secrets.username.borrow_mut() = username.to_string();
                                secrets_changed = true;
                            }
                        }
                        if ! password.is_empty() {
                            // Encrypt password with lesspass
                            let charset = lesspass::CharacterSet::All;
                            let salt = lesspass::generate_salt("lesspass.com", &username, 1);
                            let entropy = lesspass::generate_entropy(&password, &salt, lesspass::Algorithm::SHA256, 100000);
                            *secrets.password.borrow_mut() = lesspass::render_password(&entropy, charset, 16);
                            window.imp().password.set_text("");
                            secrets_changed = true;
                        }
                        if secrets_changed {
                            secrets.save().await.expect("Cannot save credentials");
                            let toast = adw::Toast::new(&gettext("Server settings saved"));
                            window.add_toast(toast);
                        }
                        window.imp().secrets_changed.set(secrets_changed);
                    }
                ));
            }
        ));
    }
}
