//
// LPTK ui window.rs
// Copyright (C) 2025 Óscar García Amor <ogarcia@connectical.com>
// Distributed under terms of the GNU GPLv3 license.
//

use std::cell::{OnceCell, RefCell};

use gettextrs::gettext;
use glib::clone;
use gtk::prelude::*;
use adw::prelude::*;
use adw::subclass::prelude::*;
use gtk::{gio, glib};

use crate::client::{Client, Secrets};
use crate::config::APP_ID;
use crate::model::Password;
use super::FragmentGenerator;
use super::FragmentPasswordList;
use super::Preferences;

mod imp {
    use super::*;

    #[derive(Debug, Default, gtk::CompositeTemplate)]
    #[template(resource = "/me/ogarcia/lptk/gtk/window.ui")]
    pub struct LptkWindow {
        // Template widgets
        #[template_child]
        pub spinner: TemplateChild<adw::Spinner>,
        #[template_child]
        pub toast_overlay: TemplateChild<adw::ToastOverlay>,
        #[template_child]
        pub stack: TemplateChild<adw::ViewStack>,
        #[template_child]
        pub fragment_generator: TemplateChild<FragmentGenerator>,
        #[template_child]
        pub fragment_password_list: TemplateChild<FragmentPasswordList>,
        pub settings: OnceCell<gio::Settings>,
        pub secrets: OnceCell<Secrets>,
        pub client: RefCell<Client>,
    }

    #[glib::object_subclass]
    impl ObjectSubclass for LptkWindow {
        const NAME: &'static str = "LptkWindow";
        type Type = super::LptkWindow;
        type ParentType = adw::ApplicationWindow;

        fn class_init(klass: &mut Self::Class) {
            klass.bind_template();
        }

        fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
            obj.init_template();
        }
    }

    impl ObjectImpl for LptkWindow {
        fn constructed(&self) {
            self.parent_constructed();

            // Load secrets from keyring, setup client, and load password list
            glib::spawn_future_local(clone!(
                #[weak(rename_to = window)]
                self,
                async move {
                    window.spinner.set_visible(true);
                    let secrets = Secrets::new();
                    let secrets = window.secrets.get_or_init(|| secrets);
                    secrets.update().await;
                    let server = secrets.server.borrow().clone();
                    let username = secrets.username.borrow().clone();
                    let password = secrets.password.borrow().clone();
                    if ! server.is_empty() && ! username.is_empty() && ! password.is_empty() {
                        // Only if credentials are stored
                        let client = Client::new(&server, &username, &password);
                        client.read_token().await;
                        *window.client.borrow_mut() = client;
                        window.obj().load_password_list();
                    } else {
                        window.spinner.set_visible(false);
                    }
                }
            ));

            // Load latest window state
            let obj = self.obj();
            obj.setup_settings();
            obj.load_window_size();
            obj.setup_actions();
            obj.setup_signals();
        }
    }

    impl WidgetImpl for LptkWindow {}

    impl WindowImpl for LptkWindow {
        // Save window state right before the window will be closed
        fn close_request(&self) -> glib::Propagation {
            // Save window size
            self.obj().save_window_size().expect("Could not save window state");
            // Allow to invoke other event handlers
            glib::Propagation::Proceed
        }
    }

    impl ApplicationWindowImpl for LptkWindow {}
    impl AdwApplicationWindowImpl for LptkWindow {}
}

glib::wrapper! {
    pub struct LptkWindow(ObjectSubclass<imp::LptkWindow>)
        @extends gtk::Widget, gtk::Window, gtk::ApplicationWindow, adw::ApplicationWindow,
        @implements gio::ActionGroup, gio::ActionMap;
}

impl LptkWindow {
    pub fn new<P: IsA<gtk::Application>>(application: &P) -> Self {
        glib::Object::builder()
            .property("application", application)
            .build()
    }

    fn setup_settings(&self) {
        // Read settings from application schema
        let settings = gio::Settings::new(APP_ID);
        self.imp().settings.set(settings).expect("Could not read settings");
    }

    #[inline(always)]
    fn settings(&self) -> &gio::Settings {
        // Return settings from self
        self.imp().settings.get().expect("Could not obtain settings")
    }

    fn load_window_size(&self) {
        // Get the window state from settings
        let width = self.settings().int("window-width");
        let height = self.settings().int("window-height");
        let maximized = self.settings().boolean("window-maximized");

        // Set the size of the window
        self.set_default_size(width, height);

        // If the window was maximized when it was closed, maximize it again
        if maximized {
            self.maximize();
        }
    }

    fn save_window_size(&self) -> Result<(), glib::BoolError> {
        // Get the size of the window
        let size = self.default_size();

        // Set the window state in `settings`
        self.settings().set_int("window-width", size.0)?;
        self.settings().set_int("window-height", size.1)?;
        self.settings().set_boolean("window-maximized", self.is_maximized())?;

        Ok(())
    }

    fn load_password_list(&self) {
        // Setup a channel to obtain password list
        let (sender, receiver) = async_channel::bounded(1);
        let client = self.imp().client.borrow().clone();
        crate::RUNTIME.spawn(clone!(
            #[strong]
            sender,
            async move {
                let password_list = client.get_passwords().await;
                sender.send(password_list).await.expect("The channel needs to be open");
            }
        ));
        glib::spawn_future_local(clone!(
            #[weak(rename_to = window)]
            self,
            async move {
                while let Ok(password_list) = receiver.recv().await {
                    window.imp().fragment_password_list.clear();
                    window.imp().spinner.set_visible(false);
                    match password_list {
                        Ok(password_list) => {
                            window.imp().fragment_password_list.populate(password_list);
                            // User can save or update passwors
                            window.imp().fragment_generator.emit_by_name::<()>("visible-save-button", &[&true]);
                        },
                        Err(err) => {
                            // If server has problems user cannot save or update passwors
                            window.imp().fragment_generator.emit_by_name::<()>("visible-save-button", &[&false]);
                            let dialog = adw::AlertDialog::builder()
                                .heading(&gettext("LessPass server error"))
                                .body(err.message())
                                .default_response("ok")
                                .build();
                            dialog.add_response("ok", &gettext("Ok"));
                            dialog.present(Some(&window));
                        }
                    }
                }
            }
        ));
    }

    fn reload(&self) {
        // Reload secrets, recreate client and get password list again
        glib::spawn_future_local(clone!(
            #[weak(rename_to = window)]
            self,
            async move {
                window.imp().spinner.set_visible(true);
                let secrets = window.imp().secrets.get().expect("Could not obtain secrets");
                secrets.update().await;
                let server = secrets.server.borrow().clone();
                let username = secrets.username.borrow().clone();
                let password = secrets.password.borrow().clone();
                if ! server.is_empty() && ! username.is_empty() && ! password.is_empty() {
                    // Only if all credentials are stored
                    let client = Client::new(&server, &username, &password);
                    *window.imp().client.borrow_mut() =  client;
                    window.load_password_list();
                } else {
                    window.imp().spinner.set_visible(false);
                }
            }
        ));
    }

    fn setup_actions(&self) {
        // Register preferences action and add to window
        let preferences_action = gio::SimpleAction::new("preferences", None);
        preferences_action.connect_activate(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, _| {
                let preferences = Preferences::new();
                // When close signal is received from settings emit a clear signal
                preferences.connect_closure("close", false, glib::closure_local!(
                    #[weak]
                    window,
                    move |_: Preferences, secrets_changed: bool| {
                        // Emit a clear signal to generator
                        window.imp().fragment_generator.emit_by_name::<()>("clear", &[&true]);
                        if secrets_changed {
                            // Reload password list
                            window.reload();
                        }
                    }
                ));
                preferences.present(Some(&window));
            }
        ));
        self.add_action(&preferences_action);

        // Register save action and add to window
        let save_action = gio::SimpleAction::new("save", None);
        save_action.connect_activate(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, _| {
                window.imp().fragment_generator.emit_by_name::<()>("save-action", &[&true]);
            }
        ));
        self.add_action(&save_action);

        // Register save action and add to window
        let delete_action = gio::SimpleAction::new("delete", None);
        delete_action.connect_activate(clone!(
            #[weak(rename_to = window)]
            self,
            move |_, _| {
                window.imp().fragment_generator.emit_by_name::<()>("delete-action", &[&true]);
            }
        ));
        self.add_action(&delete_action);
    }

    fn setup_signals(&self) {
        // When toast signal is received show it
        self.imp().fragment_generator.connect_closure("toast", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: FragmentGenerator, message: String| {
                let toast = adw::Toast::new(&message);
                window.imp().toast_overlay.add_toast(toast);
            }
        ));

        // When password selected signal is received from password list load it into generator
        self.imp().fragment_password_list.connect_closure("password-selected", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: FragmentPasswordList, password: Password| {
                window.imp().fragment_generator.load_password(password);
                window.imp().stack.set_visible_child_name("generator");
            }
        ));

        // When password save signal is received upload to backend
        self.imp().fragment_generator.connect_closure("password-save", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: FragmentGenerator, password: Password| {
                // Setup a channel to obtain password save result
                let (sender, receiver) = async_channel::bounded(1);
                window.imp().spinner.set_visible(true);
                let client = window.imp().client.borrow().clone();
                let id = password.property::<String>("id");
                let site = password.property::<String>("site");
                let login = password.property::<String>("login");
                let lower = password.property::<bool>("lower");
                let upper = password.property::<bool>("upper");
                let digits = password.property::<bool>("digits");
                let symbols = password.property::<bool>("symbols");
                let length = password.property::<u8>("length");
                let counter = password.property::<u32>("counter");
                crate::RUNTIME.spawn(clone!(
                    #[strong]
                    sender,
                    async move {
                        if id.is_empty() {
                            let save_result = client.save_password(site, login, lower, upper, digits, symbols, length, counter).await;
                            sender.send((true, save_result)).await.expect("The channel needs to be open");
                        } else {
                            let update_result = client.update_password(id, site, login, lower, upper, digits, symbols, length, counter).await;
                            sender.send((false, update_result)).await.expect("The channel needs to be open");
                        }
                    }
                ));
                glib::spawn_future_local(clone!(
                    #[weak]
                    window,
                    async move {
                        while let Ok((is_new, result)) = receiver.recv().await {
                            match result {
                                Ok(()) => {
                                    let message = if is_new {
                                        gettext("Password saved")
                                    } else {
                                        gettext("Password updated")
                                    };
                                    let toast = adw::Toast::new(&message);
                                    window.imp().toast_overlay.add_toast(toast);
                                    window.load_password_list();
                                },
                                Err(err) => {
                                    let dialog = adw::AlertDialog::builder()
                                        .heading(&gettext("LessPass server error"))
                                        .body(err.message())
                                        .default_response("ok")
                                        .build();
                                    dialog.add_response("ok", &gettext("Ok"));
                                    dialog.present(Some(&window));
                                    window.imp().spinner.set_visible(false);
                                    window.imp().fragment_generator.emit_by_name::<()>("enable-save-button", &[&true]);
                                }
                            }
                        }
                    }
                ));
            }
        ));

        // When password delete signal is received delete from backend
        self.imp().fragment_generator.connect_closure("password-delete", false, glib::closure_local!(
            #[weak(rename_to = window)]
            self,
            move |_: FragmentGenerator, id: String| {
                // Setup a channel to obtain password save result
                let (sender, receiver) = async_channel::bounded(1);
                window.imp().spinner.set_visible(true);
                let client = window.imp().client.borrow().clone();
                crate::RUNTIME.spawn(clone!(
                    #[strong]
                    sender,
                    async move {
                        let delete_result = client.delete_password(id).await;
                        sender.send(delete_result).await.expect("The channel needs to be open");
                    }
                ));
                glib::spawn_future_local(clone!(
                    #[weak]
                    window,
                    async move {
                        while let Ok(result) = receiver.recv().await {
                            match result {
                                Ok(()) => {
                                    let toast = adw::Toast::new(&gettext("Password deleted"));
                                    window.imp().toast_overlay.add_toast(toast);
                                    window.imp().fragment_generator.emit_by_name::<()>("clear", &[&true]);
                                    window.load_password_list();
                                },
                                Err(err) => {
                                    let dialog = adw::AlertDialog::builder()
                                        .heading(&gettext("LessPass server error"))
                                        .body(err.message())
                                        .default_response("ok")
                                        .build();
                                    dialog.add_response("ok", &gettext("Ok"));
                                    dialog.present(Some(&window));
                                    window.imp().spinner.set_visible(false);
                                }
                            }
                        }
                    }
                ));
            }
        ));
    }
}
